#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2023 Bootloader.  All rights reserved.
#
# This software is the confidential and proprietary information of
# Bootloader or one of its subsidiaries.  You shall not disclose this
# confidential information and shall use it only in accordance with the
# terms of the license agreement or other applicable agreement you
# entered into with Bootloader.
#
# BOOTLOADER MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE
# SUITABILITY OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT
# NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
# A PARTICULAR PURPOSE, OR NON-INFRINGEMENT.  BOOTLOADER SHALL NOT BE
# LIABLE FOR ANY LOSSES OR DAMAGES SUFFERED BY LICENSEE AS A RESULT OF
# USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.

import argparse

from majormode.perseus.constant.logging import LOGGING_LEVEL_LITERAL_STRINGS
from majormode.perseus.constant.logging import LoggingLevelLiteral
from majormode.perseus.model.date import ISO8601DateTime
from majormode.perseus.utils.logging import DEFAULT_LOGGING_FORMATTER
from majormode.perseus.utils.logging import cast_string_to_logging_level
from majormode.perseus.utils.logging import set_up_logger

from bootloader.utils.plasticscm import reporting


def parse_arguments() -> argparse.Namespace:
    """
    Convert argument strings to objects and assign them as attributes of
    the namespace.


    @return: An instance `Namespace` corresponding to the populated
        namespace.
    """
    parser = argparse.ArgumentParser(description="Plastic SCM repository activities reporting")

    parser.add_argument(
        '--server',
        dest='plasticscm_server_address',
        metavar='NAME',
        required=False,
        help="specify the Plastic SCM server to connect to"
    )

    parser.add_argument(
        '--logging-level',
        dest='logging_level',
        metavar='LEVEL',
        required=False,
        default=str(LoggingLevelLiteral.info),
        type=cast_string_to_logging_level,
        help=f"specify the logging level ({', '.join(LOGGING_LEVEL_LITERAL_STRINGS)})"
    )

    return parser.parse_args()


def run():
    arguments = parse_arguments()

    set_up_logger(
        logging_formatter=DEFAULT_LOGGING_FORMATTER,
        logging_level=arguments.logging_level
    )

    plastic_scm_version = reporting.get_cm_version()
    print(plastic_scm_version)

    available_repositories = reporting.get_available_repositories(
        server_address=arguments.plasticscm_server_address
    )

    for repository_name in available_repositories:
        changesets = reporting.get_changeset_history(
            repository_name,
            end_date=ISO8601DateTime.now(),
            include_details=True
        )




# Copyright (C) 2023 Bootloader.  All rights reserved.
#
# This software is the confidential and proprietary information of
# Bootloader or one of its subsidiaries.  You shall not disclose this
# confidential information and shall use it only in accordance with the
# terms of the license agreement or other applicable agreement you
# entered into with Bootloader.
#
# BOOTLOADER MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE
# SUITABILITY OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT
# NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
# A PARTICULAR PURPOSE, OR NON-INFRINGEMENT.  BOOTLOADER SHALL NOT BE
# LIABLE FOR ANY LOSSES OR DAMAGES SUFFERED BY LICENSEE AS A RESULT OF
# USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.

from __future__ import annotations

import re
from os import PathLike
from pathlib import Path
from typing import Any

from majormode.perseus.model.date import ISO8601DateTime
from majormode.perseus.utils import cast


class Changeset:
    """
    A changeset describes the exact differences between two successive
    versions in the version control system's repository of changes.
    Changesets are typically treated as an atomic unit, an indivisible set,
    by version control systems.
    """
    CSV_FIELD_INDEX_OWNER = 0
    CSV_FIELD_INDEX_DATE = 1
    CSV_FIELD_INDEX_CHANGESET_ID = 2

    REGEX_PATTERN_PLASTIC_SCM_TIMESTAMP = r'^(?P<month>[1-9]|1[0-9])\/(?P<day>[1-9]|[1-2][0-9]|3[01])\/(?P<year>2[0-9]{3})\s(?P<hour>[0-9]|1[0-2]):(?P<minute>[0-5][0-9]):(?P<second>[0-5][0-9])(?P<meridiem>AM|PM)$'
    REGEX_PLASTIC_SCM_TIMESTAMP = re.compile(REGEX_PATTERN_PLASTIC_SCM_TIMESTAMP)

    @classmethod
    def __cast_stringified_timestamp_to_iso8601(cls, timestamp: str) -> ISO8601DateTime:
        """
        10/18/2023

        :param timestamp:

        :return:
        """
        match = cls.REGEX_PLASTIC_SCM_TIMESTAMP.match(timestamp)
        if match is None:
            raise ValueError(f"Invalid timestamp '{timestamp}'")

        day = match.group('day')
        month = match.group('month')
        year = match.group('year')

        hours = int(match.group('hour'))
        minutes = match.group('minute')
        seconds = match.group('second')

        meridiem = match.group('meridiem')

        if meridiem == 'AM' and hours == 12:
            hours = '0'
        elif meridiem == 'PM' and hours < 12:
            hours = str(int(hours) + 12)

        stringified_iso_8601_timestamp = f'{year}-{month:>02}-{day:>02}T{hours:>02}:{minutes}:{seconds}'

        return cast.string_to_timestamp(stringified_iso_8601_timestamp)

    def __init__(
            self,
            repository_name: str,
            owner: str,
            date: ISO8601DateTime,
            changeset_id: int):
        self.__repository_name = repository_name
        self.__owner = owner
        self.__date = date
        self.__changeset_id = changeset_id
        self.__details = None

    @property
    def changeset_id(self) -> int:
        return self.__changeset_id

    @property
    def date(self) -> ISO8601DateTime:
        return self.__date

    @property
    def details(self) -> ChangesetDetail:
        return self.__details

    @details.setter
    def details(self, details: ChangesetDetail) -> None:
        if self.__details is not None:
            raise ValueError(f"The changeset {self.__changeset_id} has already details defined")
        self.__details = details

    @classmethod
    def from_csv(cls, repository_name: str, payload: Any) -> Changeset or None:
        return payload and Changeset(
            repository_name,
            payload[cls.CSV_FIELD_INDEX_OWNER],
            cls.__cast_stringified_timestamp_to_iso8601(payload[cls.CSV_FIELD_INDEX_DATE]),
            int(payload[cls.CSV_FIELD_INDEX_CHANGESET_ID])
        )

    @property
    def owner(self) -> str:
        return self.__owner

    @property
    def repository_name(self) -> str:
        return self.__repository_name


class ChangesetDetail:
    CSV_FIELD_INDEX_FILE_PATH = 0
    CSV_FIELD_INDEX_FILE_TYPE = 1
    CSV_FIELD_INDEX_FILE_SIZE = 2

    def __init__(
            self,
            file_path: PathLike,
            file_type: str,
            file_size: int
    ):
        self.__file_path = file_path
        self.__file_type = file_type
        self.__file_size = file_size

    @property
    def file_path(self) -> PathLike:
        return self.__file_path

    @property
    def file_size(self) -> int:
        return self.__file_size

    @property
    def file_type(self) -> str:
        return self.__file_type

    @classmethod
    def from_csv(cls, payload: Any) -> ChangesetDetail:
        return payload and ChangesetDetail(
            Path(payload[cls.CSV_FIELD_INDEX_FILE_PATH]),
            payload[cls.CSV_FIELD_INDEX_FILE_TYPE],
            int(payload[cls.CSV_FIELD_INDEX_FILE_SIZE])
        )



# Copyright (C) 2023 Bootloader.  All rights reserved.
#
# This software is the confidential and proprietary information of
# Bootloader or one of its subsidiaries.  You shall not disclose this
# confidential information and shall use it only in accordance with the
# terms of the license agreement or other applicable agreement you
# entered into with Bootloader.
#
# BOOTLOADER MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE
# SUITABILITY OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT
# NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
# A PARTICULAR PURPOSE, OR NON-INFRINGEMENT.  BOOTLOADER SHALL NOT BE
# LIABLE FOR ANY LOSSES OR DAMAGES SUFFERED BY LICENSEE AS A RESULT OF
# USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.

import json
import logging
import subprocess
from collections import defaultdict

from majormode.perseus.model.date import ISO8601DateTime

from bootloader.utils.plasticscm.model import Changeset, ChangesetDetail


def get_cm_version() -> str:
    command_line = ['cm', 'version']

    logging.debug(f"Running the command line '{' '.join(command_line)}'...")
    result = subprocess.run(
        command_line,
        capture_output=True,
        text=True
    )

    if result.returncode > 0:
        raise Exception(result.stderr)

    return result.stdout.strip()


def get_available_repositories(server_address: str = None) -> list[str]:
    """
    Return the list of available repositories.

    The Plastic SCM server address is also known as the "repository server
    spec" (``repserverspec``).  It can be represented with the string
    ``repserver:name:port``. For examples:

    ```text
    repserver:skull:8084
    skull:8084
    ```

    Plastic SCM call it "repository server spec", instead of just "server
    spec" for historical reasons.  Long ago, they had separate workspace
    and repository servers, and the naming survived.


    :param server_address The hostname and port of the Plastic SCM server
        to connect to and return the list of available repositories.  When
        not defined, the default server is used, i.e., the server
        configured in the configuration wizard and is written in the local
        ``client.conf`` file.


    :return: An array of the names of available repositories.
    """
    command_line = ['cm', 'repository', 'list']
    if server_address is not None:
        command_line.append(server_address)

    logging.debug(f"Running the command line '{' '.join(command_line)}'...")
    result = subprocess.run(
        command_line,
        capture_output=True,
        text=True
    )

    if result.returncode > 0:
        raise Exception(result.stderr)

    return result.stdout.strip().split('\n')


# Fetch commit history
def get_changeset_history(
        repository_name: str,
        end_date: ISO8601DateTime = None,
        include_details: bool = False,
        start_date: ISO8601DateTime = None,
) -> list[(str, str, str)]:
    """
    Return a history of changesets from the specified repository.


    :param repository_name: The name of the Plastic SCM repository to
        return the changeset history.

    :param end_date: The latest date of changesets to return.  This date
        is exclusive, so changesets that were made at this date are not
        returned.

    :param include_details: Indicate to return the details of each
        changeset.

    :param start_date: The earliest date of changesets to return.  This
        date is inclusive, so changesets that were made at this date are
        returned.


    :return: A list of tuple ``(owner, date, changeset_id)``.
    """
    command_line = ['cm', 'find', 'changeset']

    if start_date or end_date:
        filters = []
        if start_date:
            filters.append(f"date >= '{start_date}'")
        if end_date:
            filters.append(f"date < '{end_date}'")

        where_section = f"where {' and '.join(filters)}"
        command_line.append(where_section)

    command_line.extend(['on', 'repository', f"'{repository_name}'"])
    command_line.append('--format={owner};{date};{changesetid}')
    # command_line.append('--dateformat="yyyy-MM-dd HH:mm:ss"')
    command_line.append('--nototal')

    shell_command_line = ' '.join(command_line)
    logging.debug(f"Running the command line '{shell_command_line}'...")

    result = subprocess.run(
        command_line,
        capture_output=True,
        text=True
    )

    if result.returncode > 0:
        raise Exception(result.stderr)

    changeset_records = result.stdout.strip().split('\n')

    changesets = [
        Changeset.from_csv(
            repository_name,
            record.replace('"', '').split(';')
        )
        for record in changeset_records
    ]

    if include_details:
        for changeset in changesets:
            changeset.details = get_changeset_details(changeset)

    return changesets


# Fetch changed files and lines in a commit
def get_changeset_details(changeset: Changeset):
    command_line = ['cm', 'find', 'revisions']
    command_line.append(f"where changeset={changeset.changeset_id} and type!='dir'")
    command_line.extend(['on', 'repository', f"'{changeset.repository_name}'"])
    command_line.append('--format={item};{type};{size}')
    command_line.append('--nototal')

    logging.debug(f"Running the command line '{' '.join(command_line)}'...")
    result = subprocess.run(
        command_line,
        capture_output=True,
        text=True
    )

    if result.returncode > 0:
        raise Exception(result.stderr)

    changeset_detail_records = result.stdout.strip().split('\n')

    changeset_details = [
        ChangesetDetail.from_csv(record.split(';'))
        for record in changeset_detail_records
        if record != ''
    ]

    return changeset_details


# Main script starts here
if __name__ == '__main__':
    commits = get_changeset_history()
    dev_stats = defaultdict(list)

    for commit in commits:
        owner, date, changeset_id = commit.split(';')
        date = date.split(' ')[0]  # Only keep the date part, remove time
        changeset_id = int(changeset_id)

        commit_details = get_changeset_details(changeset_id)

        file_changes = []
        for detail in commit_details:
            if detail:
                data = json.loads(detail)
                file_changes.append({'file': data['Path'], 'lines_changed': data['Added'] + data['Deleted']})

        dev_stats[owner].append({'date': date, 'files': file_changes})

    # Output the stats
    for dev, commits in dev_stats.items():
        print(f"Developer: {dev}")
        for commit in commits:
            print(f"  Date: {commit['date']}")
            for file_change in commit['files']:
                print(f"    File: {file_change['file']}, Lines Changed: {file_change['lines_changed']}")

